--[[ BBConstants.lua

This file (should eventually) contain the static data for player state, buff relationships, and such

Responsibilities:
-Record the spellids & metadata for relevant buffs
-Record the interchangeable & mutually exclusive buff relationships
-Record the potential character roles
-Record the potential 

Dependencies:
None, yay!

Architecture:
Built as a Lua list, with a custom metatable, ala "setmetatable(instance_of, {__index = master_object})"
]]

-- Namespace
BBConstants = {}

-- Meta table for Cataclysm
BBConstants.mt_cata = {}
BBConstants.mt_cata.__index = function (table, key)
	return BBConstants.cataclysm[key]
end

-- Meta table for Wrath of the Lich King
BBConstants.mt_wotlk = {}
BBConstants.mt_wotlk.__index = function (table, key)
	return BBConstants.wotlk[key]
end

-- Constructor
function BBConstants.new (clientVersion)
	local newInstance = {}
	
	-- if: Wrath of the Lich King
	if 30300 == clientVersion then
	
		-- default to relevant constants
		setmetatable(newInstance, BBConstants.mt_wotlk)

	-- elseif: Cataclysm
	elseif 40000 == clientVersion then
	
		-- default to relevant constants
		setmetatable(newInstance, BBConstants.mt_cata)
	end

	return newInstance	
end

BBConstants.wotlk = {
	SpecSpells = {
		Primary = true,
		Secondary = true,
	},
	TalentGroup = {
		None = -1, -- character doesn't have dual spec
		Unknown = 0, -- character hasn't been scanned yet
		Primary = 1, -- primary spec active
		Secondary = 2, -- secondary spec active
	},
		
	Confidence = {
		Certain = 5, -- can provide best (as can others), all others committed
		Likely = 4, -- can provide best (as can others), no coverage
		Possible = 3, -- can provide best, slot coverage
		Optional = 2, -- can't provide best, no slot coverage
		Unlikely = 1, -- player coverage (not us); or can't provide best, slot coverage
		None = 0,
	},
	Types = {
		Player = 0,
		WorldObject = 1,
		NPC = 3, -- includes temporary pets/guardians
		CombatPet = 4,
		Vehicle = 5,
		
	},
	Classes = {
		Warrior = "WARRIOR",
		Deathknight = "DEATHKNIGHT",
		Paladin = "PALADIN",
		Shaman = "SHAMAN",
		Hunter = "HUNTER",
		Rogue = "ROGUE",
		Druid = "DRUID",
		Warlock = "WARLOCK",
		Priest = "PRIEST",
		Mage = "MAGE",
		Raid = "RAID",
	},
	Roles = {
		None = "No Role",
		ManaUser = "Caster",
		NoMana = "Melee DPS",
		Unknown = "Unknown Role",
		Tank = "Tank",
		PureMelee = "Melee DPS",
		MeleeMana = "Melee DPS (mana user)",
		Caster = "Caster",
		MeleeMana_Low = "Low level Melee DPS (mana user)",
		Healer = "Healer",
	},
	Totems = {
		["FIRE_TOTEMS"] = 133,
		["EARTH_TOTEMS"] = 134,
		["WATER_TOTEMS"] = 135,
		["AIR_TOTEMS"] = 136,
	},
	Buffed = {
		None,
		Partial,
		Full
	},
	-- TODO:  Figure out a better name for this, i.e. 'How improved this buff is'
	Buffs = {
		None = 1, -- can't provide
		ProfessionStats = 2, -- Leatherworking Kings
		Basic = 3, -- class inherent (shaman/warrior)
		BasicLong = 4, -- class inherent (paladin): wisdom, might, kings, sanctuary
		MinorTalented = 5, -- partial talent investment, warrior (1,2 /5), shaman (1/3)
		PartialTalented = 6, -- partial talent investment, druid (1/2), priest (1/2)
		PartialTalentedLong = 7, -- partial talent investment, paladin (1/2)
		MajorTalented = 8, -- partial talent investment, warrior (3,4 /5), shaman (2/3)
		FullTalented = 8, -- 3/3 shaman totems, 5/5 warrior shouts, druid (2/2), priest (2/2)
		FullTalentedLong = 9, -- 2/2 paladin wisdom, 2/2 Paladin might
		Awesome = 10, -- way better than anything we could do
	},
	Scope = {
		Individual = 1,
		Class = 2,
		Raid = 3,
		Self = 4,
	},
	MaxBuffSlots = 40,
	MaxPlayers = 40,
}

BBConstants.wotlk.Forms = { -- warrior stances are dum-dums for now, until blizzard makes them an aura, OR, forms are checked on spellcast instead of aura scan (which is cludgy)
		[71] = BBConstants.wotlk.Roles.Tank, -- defensive stance
		[2457] = BBConstants.wotlk.Roles.PureMelee, -- battle stance
		[2458] = BBConstants.wotlk.Roles.PureMelee, -- Berserker stance
		[25780] = BBConstants.wotlk.Roles.Tank, -- righteous fury
		[48263] = BBConstants.wotlk.Roles.Tank, -- frost presence
		[48265] = BBConstants.wotlk.Roles.PureMelee, -- unholy presence
		[48266] = BBConstants.wotlk.Roles.PureMelee, -- blood presence
		[5487] = BBConstants.wotlk.Roles.Tank, -- Bear Form
		[9634] = BBConstants.wotlk.Roles.Tank, -- Dire Bear Form
		[768] = BBConstants.wotlk.Roles.PureMelee, -- Cat Form
}

BBConstants.wotlk.BuffSlots = {
	["BLESSINGS_BASIC"] = {
		[19742] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 14, Scope = BBConstants.wotlk.Scope.Individual}, -- wisdom
		[25894] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 54, Scope = BBConstants.wotlk.Scope.Class}, -- greater wisdom
		[19740] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 4, Scope = BBConstants.wotlk.Scope.Individual}, -- might
		[25782] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 52, Scope = BBConstants.wotlk.Scope.Class}, -- greater might
		[20217] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 20, Scope = BBConstants.wotlk.Scope.Individual}, -- kings
		[25898] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 60, Scope = BBConstants.wotlk.Scope.Class}, -- greater kings
	},
	["BLESSINGS"] = {
		[19742] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 14, Scope = BBConstants.wotlk.Scope.Individual}, -- wisdom
		[25894] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 54, Scope = BBConstants.wotlk.Scope.Class}, -- greater wisdom
		[19740] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 4, Scope = BBConstants.wotlk.Scope.Individual}, -- might
		[25782] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 52, Scope = BBConstants.wotlk.Scope.Class}, -- greater might
		[20217] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 20, Scope = BBConstants.wotlk.Scope.Individual}, -- kings
		[25898] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 60, Scope = BBConstants.wotlk.Scope.Class}, -- greater kings
		[20911] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 30, Scope = BBConstants.wotlk.Scope.Individual}, -- sanctuary
		[25899] = {Score = BBConstants.wotlk.Buffs.BasicLong, Target = true, MinLevel = 60, Scope = BBConstants.wotlk.Scope.Class}, -- greater sanctuary
	},
	["SEALS"] = {
		[53736] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 66, Scope = BBConstants.wotlk.Scope.Self}, -- seal of corruption
		[20164] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 22, Scope = BBConstants.wotlk.Scope.Self}, -- seal of justice
		[20165] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 30, Scope = BBConstants.wotlk.Scope.Self}, -- seal of light
		[20154] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 1, Scope = BBConstants.wotlk.Scope.Self}, -- seal of righteousness
		[31801] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 64, Scope = BBConstants.wotlk.Scope.Self}, -- seal of vengeance
		[20166] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 38, Scope = BBConstants.wotlk.Scope.Self}, -- seal of wisdom
		[20375] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 20, Scope = BBConstants.wotlk.Scope.Self}, -- seal of command
	},
	["SHOUTS"] = {
		[469] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 68, Scope = BBConstants.wotlk.Scope.Raid}, -- commanding
		[6673] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 1, Scope = BBConstants.wotlk.Scope.Raid}, -- battle
	},
	["ICY_TALONS"] = {
		[55610] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 36, Scope = BBConstants.wotlk.Scope.Raid}, -- improved icy talons
	},
	["HORNS"] = {
		[57330] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 65, Scope = BBConstants.wotlk.Scope.Raid}, -- horn of winter
	},
	["WATER_TOTEMS"] = {
		[5677] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 26, Scope = BBConstants.wotlk.Scope.Raid}, -- mana spring
		[5672] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 20, Scope = BBConstants.wotlk.Scope.Raid}, -- healing stream
	},
	["EARTH_TOTEMS"] = {
		[8072] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 4, Scope = BBConstants.wotlk.Scope.Raid}, -- stoneskin
		[8076] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 10, Scope = BBConstants.wotlk.Scope.Raid}, -- strength of earth
	},
	["FIRE_TOTEMS_BASIC"] = {
		[52109] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 28, Scope = BBConstants.wotlk.Scope.Raid}, -- flametongue
		[3599] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 10, Scope = BBConstants.wotlk.Scope.Raid}, -- searing totem
	},
	["FIRE_TOTEMS"] = {
		[52109] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 28, Scope = BBConstants.wotlk.Scope.Raid}, -- flametongue
		[3599] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 10, Scope = BBConstants.wotlk.Scope.Raid}, -- searing totem
		[57658] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 50, Scope = BBConstants.wotlk.Scope.Raid}, -- wrath
	},
	["AIR_TOTEMS"] = {
		[8515] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 32, Scope = BBConstants.wotlk.Scope.Raid}, -- windfury
		[2895] = {Score = BBConstants.wotlk.Buffs.Basic, Target = false, MinLevel = 64, Scope = BBConstants.wotlk.Scope.Raid}, -- wrath of air
	},
	["FORTITUDE"] = {
		[1243] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 1, Scope = BBConstants.wotlk.Scope.Individual}, -- Power Word: Fortitude 8
		[21562] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 48, Scope = BBConstants.wotlk.Scope.Raid}, -- Prayer of Fortitude 4
	},
	["AI"] = {
		[61024] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 80, Scope = BBConstants.wotlk.Scope.Individual}, -- Dalaran Intellect
		[61316] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 80, Scope = BBConstants.wotlk.Scope.Raid}, -- Dalaran Brilliance
		[1459] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 1, Scope = BBConstants.wotlk.Scope.Individual}, -- Arcane Intellect 7 
		[23028] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 56, Scope = BBConstants.wotlk.Scope.Raid}, -- Arcane Brilliance 3
	},
	["WILD"] = {
		[1126] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 1, Scope = BBConstants.wotlk.Scope.Individual}, -- Mark of the Wild 9
		[21849] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 50, Scope = BBConstants.wotlk.Scope.Raid}, -- Gift of the Wild 4
	},
	["WILD_NPC"] = {
		[46119] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 70, Scope = BBConstants.wotlk.Scope.Individual}, -- Blessing of D E H T A
	},
	["THORNS"] = {
		[467] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 6, Scope = BBConstants.wotlk.Scope.Individual}, -- Thorns 8
	},
	["SPIRIT"] = {
		[14752] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 30, Scope = BBConstants.wotlk.Scope.Individual}, -- Divine Spirit 6
		[27681] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 60, Scope = BBConstants.wotlk.Scope.Raid}, -- Prayer of Spirit 3
	},
	["PRIEST_INNER"] = {
		[588] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 12, Scope = BBConstants.wotlk.Scope.Self}, -- Inner Fire
	},
	["PRIEST_PROTECTION"] = {
		[976] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 30, Scope = BBConstants.wotlk.Scope.Individual}, -- Shadow Protection
		[27683] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 56, Scope = BBConstants.wotlk.Scope.Raid}, -- Prayer of Shadow Protection
	},
	["EMBRACE"] = {
		[15286] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 30, Scope = BBConstants.wotlk.Scope.Self}, -- Inner Fire
	},
	["MAGE_ARMOR"] = {
		[6117] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 64, Scope = BBConstants.wotlk.Scope.Self}, -- Mage Armor
		[168] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 1, Scope = BBConstants.wotlk.Scope.Self}, -- Frost Armor
		[7302] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 30, Scope = BBConstants.wotlk.Scope.Self}, -- Ice Armor
		[30482] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 62, Scope = BBConstants.wotlk.Scope.Self}, -- Molten Armor
	},
	["WARLOCK_ARMOR"] = {
		[687] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 1, Scope = BBConstants.wotlk.Scope.Self}, -- Demon Skin
		[706] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 20, Scope = BBConstants.wotlk.Scope.Self}, -- Demon Armor
		[28176] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 62, Scope = BBConstants.wotlk.Scope.Self}, -- Fel Armor
	},
	["RIGHTEOUS_FURY"] = {
		[25780] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 20, Scope = BBConstants.wotlk.Scope.Self}, -- Righteous Fury
	},
	["PRIEST_FORM"] = {
		[15473] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 40, Scope = BBConstants.wotlk.Scope.Self}, -- Shadow Form
	},

	["SHAPESHIFT"] = {
		[24858] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 40, Scope = BBConstants.wotlk.Scope.Self}, -- Owl Form
		[33891] = {Score = BBConstants.wotlk.Buffs.Basic, Target = true, MinLevel = 50, Scope = BBConstants.wotlk.Scope.Self}, -- Tree Form
	},

	["PROFESSION_STATS"] = {
		[69378] = {Score = BBConstants.wotlk.Buffs.ProfessionStats, Scope = BBConstants.wotlk.Scope.Raid}, -- forgotten kings
	},
}


BBConstants.wotlk.GreaterLookup = {
	["BLESSINGS"] = {
		[20217] = 25898, -- Kings
		[20911] = 25899, -- Sanctuary
		[19838] = 25782, -- Might 1
		[25291] = 25916, -- Might 2
		[27140] = 27141, -- Might 3
		[48931] = 48933, -- Might 4
		[48932] = 48934, -- Might 5
		[19854] = 25894, -- Wisdom 1
		[25290] = 25918, -- Wisdom 2
		[27142] = 27143, -- Wisdom 3
		[48935] = 48937, -- Wisdom 4
		[48936] = 48938, -- Wisdom 5
	},
	["FORTITUDE"] = {
		[10937] = 21562, -- Prayer of Fortitude 1
		[10938] = 21564, -- Prayer of Fortitude 2
		[25389] = 25392, -- Prayer of Fortitude 3
		[48161] = 48162, -- Prayer of Fortitude 4
	},
	["AI"] = {
		[10157] = 23028, -- Arcane Brilliance 1
		[27126] = 27127, -- Arcane Brilliance 2
		[42995] = 43002, -- Arcane Brilliance 3
	},
	["WILD"] = {
		[9884] = 21849, -- Gift of the Wild 1
		[9885] = 21850, -- Gift of the Wild 2
		[26990] = 26991, -- Gift of the Wild 3
		[48469] = 48470, -- Gift of the Wild 4
	},
	["SPIRIT"] = {
		[27841] = 27681, -- Prayer of Spirit 1
		[25312] = 32999, -- Prayer of Spirit 2
		[48073] = 48074, -- Prayer of Spirit 3
	},
	["PRIEST_PROTECTION"] = {
		[976] = 27683, -- Prayer of Spirit 1
		[25433] = 39374, -- Prayer of Spirit 2
		[48169] = 48170, -- Prayer of Spirit 3
	},
}

-- TODO:  add 'strength';  level is close, but not a perfect guidline (especially with MP5)
BBConstants.wotlk.DownRankLookup = {
	["SHADOW_RESIST"] = {
		[976] = {minspellid=976, worstSpellID,level=30, slot="PRIEST_PROTECTION"},
		[10957] = {minspellid=976, worstSpellID,level=42, slot="PRIEST_PROTECTION"},
		[10958] = {minspellid=976, worstSpellID,level=56, slot="PRIEST_PROTECTION"},
		[27683] = {minspellid=27683, worstSpellID,level=56, slot="PRIEST_PROTECTION"},
		[25433] = {minspellid=976, worstSpellID,level=68, slot="PRIEST_PROTECTION"},
		[39374] = {minspellid=27683, worstSpellID,level=70, slot="PRIEST_PROTECTION"},
		[48169] = {minspellid=976, worstSpellID,level=76, slot="PRIEST_PROTECTION"},
		[48170] = {minspellid=27683, worstSpellID,level=77, slot="PRIEST_PROTECTION"},
	},
	["AP"] = {
		[6673] = {minspellid=6673, worstSpellID,level=1, slot="SHOUTS"},
		[19740] = {minspellid=19740,level=4, slot="BLESSINGS"},
		[5242] = {minspellid=6673,level=12, slot="SHOUTS"},
		[19834] = {minspellid=19740,level=12, slot="BLESSINGS"},
		[6192] = {minspellid=6673,level=22, slot="SHOUTS"},
		[19835] = {minspellid=19740,level=22, slot="BLESSINGS"},
		[11549] = {minspellid=6673,level=32, slot="SHOUTS"},
		[19836] = {minspellid=19740,level=32, slot="BLESSINGS"},
		[11550] = {minspellid=6673,level=42, slot="SHOUTS"},
		[19837] = {minspellid=19740,level=42, slot="BLESSINGS"},
		[11551] = {minspellid=6673,level=52, slot="SHOUTS"},
		[19838] = {minspellid=19740,level=52, slot="BLESSINGS"},
		[25782] = {minspellid=25782,level=52, slot="BLESSINGS"},
		[25289] = {minspellid=6673,level=60, slot="SHOUTS"},
		[25291] = {minspellid=19740,level=60, slot="BLESSINGS"},
		[25916] = {minspellid=25782,level=60, slot="BLESSINGS"},
		[2048] = {minspellid=6673,level=69, slot="SHOUTS"},
		[27140] = {minspellid=19740,level=70, slot="BLESSINGS"},
		[27141] = {minspellid=25782,level=70, slot="BLESSINGS"},
		[48931] = {minspellid=19740,level=73, slot="BLESSINGS"},
		[48933] = {minspellid=25782,level=73, slot="BLESSINGS"},
		[47436] = {minspellid=6673,level=78, slot="SHOUTS"},
		[48932] = {minspellid=19740,level=79, slot="BLESSINGS"},
		[48934] = {minspellid=25782,level=79, slot="BLESSINGS"},
	},
	["MP5"] = {
		[19742] = {minspellid=19742,level=14, slot="BLESSINGS"},
		[19850] = {minspellid=19742,level=24, slot="BLESSINGS"},
		[5677] = {minspellid=5677,level=26, totemspellid=5675, slot="WATER_TOTEMS"},
		[5675] = {minspellid=5675,level=26, slot="WATER_TOTEMS"},
		[19852] = {minspellid=19742,level=34, slot="BLESSINGS"},
		[10491] = {minspellid=5677,level=36, totemspellid=10495, slot="WATER_TOTEMS"},
		[10495] = {minspellid=5675,level=36, slot="WATER_TOTEMS"},
		[19853] = {minspellid=19742,level=44, slot="BLESSINGS"},
		[10493] = {minspellid=5677,level=46, totemspellid=10496, slot="WATER_TOTEMS"},
		[10496] = {minspellid=5675,level=46, slot="WATER_TOTEMS"},
		[19854] = {minspellid=19742,level=54, slot="BLESSINGS"},
		[25894] = {minspellid=25894,level=54, slot="BLESSINGS"},
		[10494] = {minspellid=5677,level=56, totemspellid=10497, slot="WATER_TOTEMS"},
		[10497] = {minspellid=5675,level=56, slot="WATER_TOTEMS"},
		[25290] = {minspellid=19742,level=60, slot="BLESSINGS"},
		[25918] = {minspellid=25894,level=60, slot="BLESSINGS"},
		[25569] = {minspellid=5677,level=65, totemspellid=25570, slot="WATER_TOTEMS"},
		[25570] = {minspellid=5675,level=65, slot="WATER_TOTEMS"},
		[27142] = {minspellid=19742,level=65, slot="BLESSINGS"},
		[27143] = {minspellid=25894,level=65, slot="BLESSINGS"},
		[58775] = {minspellid=5677,level=71, totemspellid=58771, slot="WATER_TOTEMS"},
		[58771] = {minspellid=5675,level=71, slot="WATER_TOTEMS"},
		[48935] = {minspellid=19742,level=71, slot="BLESSINGS"},
		[48937] = {minspellid=5675,level=71, slot="BLESSINGS"},
		[58776] = {minspellid=5675,level=76, totemspellid=58773, slot="WATER_TOTEMS"},
		[58773] = {minspellid=5675,level=76, slot="WATER_TOTEMS"},
		[48936] = {minspellid=5675,level=77, slot="BLESSINGS"},
		[48938] = {minspellid=25894,level=77, slot="BLESSINGS"},
		[58777] = {minspellid=5677,level=80, totemspellid=58774, slot="WATER_TOTEMS"},
		[58774] = {minspellid=5675,level=80, slot="WATER_TOTEMS"},
	},
	["HP5"] = {
		[5672] = {minspellid=5672,level=20, totemspellid=5394, slot="WATER_TOTEMS"},
		[5394] = {minspellid=5394,level=20, slot="WATER_TOTEMS"},
		[6371] = {minspellid=5672,level=30, totemspellid=6375, slot="WATER_TOTEMS"},
		[6375] = {minspellid=5394,level=30, slot="WATER_TOTEMS"},
		[6372] = {minspellid=5672,level=40, totemspellid=6377, slot="WATER_TOTEMS"},
		[6377] = {minspellid=5394,level=40, slot="WATER_TOTEMS"},
		[10460] = {minspellid=5672,level=50, totemspellid=10462, slot="WATER_TOTEMS"},
		[10462] = {minspellid=5394,level=50, slot="WATER_TOTEMS"},
		[10461] = {minspellid=5672,level=60, totemspellid=10463, slot="WATER_TOTEMS"},
		[10463] = {minspellid=5394,level=60, slot="WATER_TOTEMS"},
		[25566] = {minspellid=5672,level=69, totemspellid=25567, slot="WATER_TOTEMS"},
		[25567] = {minspellid=5394,level=69, slot="WATER_TOTEMS"},
		[58763] = {minspellid=5672,level=71, totemspellid=58755, slot="WATER_TOTEMS"},
		[58755] = {minspellid=5394,level=71, slot="WATER_TOTEMS"},
		[58764] = {minspellid=5672,level=76, totemspellid=58756, slot="WATER_TOTEMS"},
		[58756] = {minspellid=5394,level=76, slot="WATER_TOTEMS"},
		[65994] = {minspellid=5672,level=80, totemspellid=58757, slot="WATER_TOTEMS"},
		[58757] = {minspellid=5394,level=80, slot="WATER_TOTEMS"},
	},
	["ARMOR"] = {
		[8072] = {minspellid=8072,level=4, totemspellid=8071, slot="EARTH_TOTEMS"},
		[8071] = {minspellid=8071,level=4, slot="EARTH_TOTEMS"},
		[8156] = {minspellid=8072,level=14, totemspellid=8154, slot="EARTH_TOTEMS"},
		[8154] = {minspellid=8071,level=14, slot="EARTH_TOTEMS"},
		[8157] = {minspellid=8072,level=24, totemspellid=8155, slot="EARTH_TOTEMS"},
		[8155] = {minspellid=8071,level=24, slot="EARTH_TOTEMS"},
		[10403] = {minspellid=8072,level=34, totemspellid=10406, slot="EARTH_TOTEMS"},
		[10406] = {minspellid=8071,level=34, slot="EARTH_TOTEMS"},
		[10404] = {minspellid=8072,level=44, totemspellid=10407, slot="EARTH_TOTEMS"},
		[10407] = {minspellid=8071,level=44, slot="EARTH_TOTEMS"},
		[10405] = {minspellid=8072,level=54, totemspellid=10408, slot="EARTH_TOTEMS"},
		[10408] = {minspellid=8071,level=54, slot="EARTH_TOTEMS"},
		[25506] = {minspellid=8072,level=63, totemspellid=25508, slot="EARTH_TOTEMS"},
		[25508] = {minspellid=8071,level=63, slot="EARTH_TOTEMS"},
		[25507] = {minspellid=8072,level=70, totemspellid=25509, slot="EARTH_TOTEMS"},
		[25509] = {minspellid=8071,level=70, slot="EARTH_TOTEMS"},
		[58752] = {minspellid=8072,level=75, totemspellid=58751, slot="EARTH_TOTEMS"},
		[58751] = {minspellid=8071,level=75, slot="EARTH_TOTEMS"},
		[58754] = {minspellid=8072,level=78, totemspellid=58753, slot="EARTH_TOTEMS"},
		[58753] = {minspellid=8071,level=78, slot="EARTH_TOTEMS"},
	},
	["STRENGTH_AGILITY"] = {
		[8076] = {minspellid=8076,level=10, totemspellid=8075, slot="EARTH_TOTEMS"},
		[8075] = {minspellid=8075,level=10, slot="EARTH_TOTEMS"},
		[8162] = {minspellid=8076,level=24, totemspellid=8160, slot="EARTH_TOTEMS"},
		[8160] = {minspellid=8075,level=24, slot="EARTH_TOTEMS"},
		[8163] = {minspellid=8076,level=38, totemspellid=8161, slot="EARTH_TOTEMS"},
		[8161] = {minspellid=8075,level=38, slot="EARTH_TOTEMS"},
		[10441] = {minspellid=8076,level=52, totemspellid=10442, slot="EARTH_TOTEMS"},
		[10442] = {minspellid=8075,level=52, slot="EARTH_TOTEMS"},
		[25362] = {minspellid=8076,level=60, totemspellid=25361, slot="EARTH_TOTEMS"},
		[25361] = {minspellid=8075,level=60, slot="EARTH_TOTEMS"},
		[25527] = {minspellid=8076,level=65, totemspellid=25528, slot="EARTH_TOTEMS"},
		[25528] = {minspellid=8075,level=65, slot="EARTH_TOTEMS"},
		[57330] = {minspellid=57330,level=65, slot="HORNS"},
		[57621] = {minspellid=8076,level=75, totemspellid=57622, slot="EARTH_TOTEMS"},
		[57622] = {minspellid=8075,level=75, slot="EARTH_TOTEMS"},
		[57623] = {minspellid=57330,level=75, slot="HORNS"},
		[58646] = {minspellid=8076,level=80, totemspellid=58643, slot="EARTH_TOTEMS"},
		[58643] = {minspellid=8075,level=80, slot="EARTH_TOTEMS"},
	},
	["SOLO_FIRE_TOTEM"] = {
		[3599] = {minspellid=3599,level=10, totemspellid=3599, slot="FIRE_TOTEMS"},
		[3599] = {minspellid=3599,level=10, slot="FIRE_TOTEMS"},
		[6363] = {minspellid=3599,level=20, totemspellid=6363, slot="FIRE_TOTEMS"},
		[6363] = {minspellid=3599,level=20, slot="FIRE_TOTEMS"},
		[6364] = {minspellid=3599,level=30, totemspellid=6364, slot="FIRE_TOTEMS"},
		[6364] = {minspellid=3599,level=30, slot="FIRE_TOTEMS"},
		[6365] = {minspellid=3599,level=40, totemspellid=6365, slot="FIRE_TOTEMS"},
		[6365] = {minspellid=3599,level=40, slot="FIRE_TOTEMS"},
		[10437] = {minspellid=3599,level=50, totemspellid=10437, slot="FIRE_TOTEMS"},
		[10437] = {minspellid=3599,level=50, slot="FIRE_TOTEMS"},
		[10438] = {minspellid=3599,level=60, totemspellid=10438, slot="FIRE_TOTEMS"},
		[10438] = {minspellid=3599,level=60, slot="FIRE_TOTEMS"},
		[25533] = {minspellid=3599,level=69, totemspellid=25533, slot="FIRE_TOTEMS"},
		[25533] = {minspellid=3599,level=69, slot="FIRE_TOTEMS"},
		[58699] = {minspellid=3599,level=71, totemspellid=58699, slot="FIRE_TOTEMS"},
		[58699] = {minspellid=3599,level=71, slot="FIRE_TOTEMS"},
		[58703] = {minspellid=3599,level=75, totemspellid=58703, slot="FIRE_TOTEMS"},
		[58703] = {minspellid=3599,level=75, slot="FIRE_TOTEMS"},
		[58704] = {minspellid=3599,level=80, totemspellid=58704, slot="FIRE_TOTEMS"},
		[58704] = {minspellid=3599,level=80, slot="FIRE_TOTEMS"},
	},
	["FLAT_SPELL"] = {
		[52109] = {minspellid=52109,level=28, totemspellid=8227, slot="FIRE_TOTEMS"},
		[8227] = {minspellid=8227,level=28, slot="FIRE_TOTEMS"},
		[52110] = {minspellid=52109,level=38, totemspellid=8249, slot="FIRE_TOTEMS"},
		[8249] = {minspellid=8227,level=38, slot="FIRE_TOTEMS"},
		[52111] = {minspellid=52109,level=48, totemspellid=10526, slot="FIRE_TOTEMS"},
		[10526] = {minspellid=8227,level=48, slot="FIRE_TOTEMS"},
		[52112] = {minspellid=52109,level=58, totemspellid=16387, slot="FIRE_TOTEMS"},
		[16387] = {minspellid=8227,level=58, slot="FIRE_TOTEMS"},
		[52113] = {minspellid=52109,level=67, totemspellid=25557, slot="FIRE_TOTEMS"},
		[25557] = {minspellid=8227,level=67, slot="FIRE_TOTEMS"},
		[58651] = {minspellid=52109,level=71, totemspellid=58649, slot="FIRE_TOTEMS"},
		[58649] = {minspellid=8227,level=71, slot="FIRE_TOTEMS"},
		[58654] = {minspellid=52109,level=75, totemspellid=58652, slot="FIRE_TOTEMS"},
		[58652] = {minspellid=8227,level=75, slot="FIRE_TOTEMS"},
		[58655] = {minspellid=52109,level=80, totemspellid=58656, slot="FIRE_TOTEMS"},
		[58656] = {minspellid=8227,level=80, slot="FIRE_TOTEMS"},
	},
	["FLAT_SPELL_CRIT"] = {
		[57658] = {minspellid=57658,level=50, totemspellid=30706, slot="FIRE_TOTEMS"},
		[30706] = {minspellid=30706,level=50, slot="FIRE_TOTEMS"},
		[57660] = {minspellid=57658,level=60, totemspellid=57720, slot="FIRE_TOTEMS"},
		[57720] = {minspellid=30706,level=60, slot="FIRE_TOTEMS"},
		[57662] = {minspellid=57658,level=70, totemspellid=57721, slot="FIRE_TOTEMS"},
		[57721] = {minspellid=30706,level=70, slot="FIRE_TOTEMS"},
		[57663] = {minspellid=57658,level=80, totemspellid=57722, slot="FIRE_TOTEMS"},
		[57722] = {minspellid=30706,level=80, slot="FIRE_TOTEMS"},
	},
	["MELEE_HASTE"] = {
		[8515] = {minspellid=8515,level=32, totemspellid=8512, slot="AIR_TOTEMS"},
		[8512] = {minspellid=8512,level=32, slot="AIR_TOTEMS"},
		[55610] = {minspellid=55610,level=36, slot="ICY_TALONS"},
	},
	["SPELL_HASTE"] = {
		[2895] = {minspellid=2895,level=64, totemspellid=3738, slot="AIR_TOTEMS"},
		[3738] = {minspellid=3738,level=64, slot="AIR_TOTEMS"},
	},
	["STATS"] = {
		[20217] = {minspellid=20217,level=20, slot="BLESSINGS"},
		[25898] = {minspellid=25898,level=60, slot="BLESSINGS"},
	},
	["TANK"] = {
		[20911] = {minspellid=20911,level=30, slot="BLESSINGS"},
		[25899] = {minspellid=25899,level=60, slot="BLESSINGS"},
	},
	["HP"] = {
		[469] = {minspellid=469,level=68, slot="SHOUTS"},
		[47439] = {minspellid=469,level=74, slot="SHOUTS"},
		[47440] = {minspellid=469,level=80, slot="SHOUTS"},
	},
	["STAMINA"] = {
		[1243] = {minspellid=1243,level=1, slot="FORTITUDE"}, -- Power Word: Fortitude 1
		[1244] = {minspellid=1243,level=12, slot="FORTITUDE"}, -- Power Word: Fortitude 2
		[1245] = {minspellid=1243,level=24, slot="FORTITUDE"}, -- Power Word: Fortitude 3
		[2791] = {minspellid=1243,level=36, slot="FORTITUDE"}, -- Power Word: Fortitude 4
		[10937] = {minspellid=1243,level=48, slot="FORTITUDE"}, -- Power Word: Fortitude 5
		[21562] = {minspellid=21562,level=48, slot="FORTITUDE"}, -- Prayer of Fortitude 1
		[10938] = {minspellid=1243,level=60, slot="FORTITUDE"}, -- Power Word: Fortitude 6
		[21564] = {minspellid=21562,level=60, slot="FORTITUDE"}, -- Prayer of Fortitude 2
		[25389] = {minspellid=1243,level=70, slot="FORTITUDE"}, -- Power Word: Fortitude 7
		[25392] = {minspellid=21562,level=70, slot="FORTITUDE"}, -- Prayer of Fortitude 3
		[48161] = {minspellid=1243,level=80, slot="FORTITUDE"}, -- Power Word: Fortitude 8
		[48162] = {minspellid=21562,level=80, slot="FORTITUDE"}, -- Prayer of Fortitude 4
	},
	["INTELLECT"] = {
		[1459] = {minspellid=1459,level=1, slot="AI"}, -- Arcane Intellect 1
		[1460] = {minspellid=1459,level=14, slot="AI"}, -- Arcane Intellect 2
		[1461] = {minspellid=1459,level=28, slot="AI"}, -- Arcane Intellect 3
		[10156] = {minspellid=1459,level=42, slot="AI"}, -- Arcane Intellect 4
		[10157] = {minspellid=1459,level=56, slot="AI"}, -- Arcane Intellect 5
		[23028] = {minspellid=23028,level=56, slot="AI"}, -- Arcane Brilliance 1
		[27126] = {minspellid=1459,level=70, slot="AI"}, -- Arcane Intellect 6
		[27127] = {minspellid=23028,level=70, slot="AI"}, -- Arcane Brilliance 2
		[42995] = {minspellid=1459,level=80, slot="AI"}, -- Arcane Intellect 7 
		[43002] = {minspellid=23028,level=80, slot="AI"}, -- Arcane Brilliance 3
		[61024] = {minspellid=61024,level=80, slot="AI"}, -- Dalaran Intellect
		[61316] = {minspellid=61316,level=80, slot="AI"}, -- Dalaran Brilliance
	},
	["WILD"] = {
		[1126] = {minspellid=1126,level=1, slot="WILD"}, -- Mark of the Wild 1
		[5232] = {minspellid=1126,level=10, slot="WILD"}, -- Mark of the Wild 2
		[6756] = {minspellid=1126,level=20, slot="WILD"}, -- Mark of the Wild 3
		[5234] = {minspellid=1126,level=30, slot="WILD"}, -- Mark of the Wild 4
		[8907] = {minspellid=1126,level=40, slot="WILD"}, -- Mark of the Wild 5
		[9884] = {minspellid=1126,level=50, slot="WILD"}, -- Mark of the Wild 6
		[21849] = {minspellid=21849,level=50, slot="WILD"}, -- Gift of the Wild 1
		[9885] = {minspellid=1126,level=60, slot="WILD"}, -- Mark of the Wild 7
		[21850] = {minspellid=21849,level=60, slot="WILD"}, -- Gift of the Wild 2
		[26990] = {minspellid=1126,level=70, slot="WILD"}, -- Mark of the Wild 8
		[26991] = {minspellid=21849,level=70, slot="WILD"}, -- Gift of the Wild 3
		[48469] = {minspellid=1126,level=80, slot="WILD"}, -- Mark of the Wild 9
		[48470] = {minspellid=21849,level=80, slot="WILD"}, -- Gift of the Wild 4
		
		[46119] = {minspellid=46119,level=70, slot="WILD_NPC"}, -- Blessing of D E H T A
	},
	["THORNS"] = {
		[467] = {minspellid=467,level=6, slot="THORNS"}, -- Thorns 1
		[782] = {minspellid=467,level=14, slot="THORNS"}, -- Thorns 2
		[1075] = {minspellid=467,level=24, slot="THORNS"}, -- Thorns 3
		[8914] = {minspellid=467,level=34, slot="THORNS"}, -- Thorns 4
		[9756] = {minspellid=467,level=44, slot="THORNS"}, -- Thorns 5
		[9910] = {minspellid=467,level=54, slot="THORNS"}, -- Thorns 6
		[26992] = {minspellid=467,level=64, slot="THORNS"}, -- Thorns 7
		[53307] = {minspellid=467,level=74, slot="THORNS"}, -- Thorns 8
	},
	["SPIRIT"] = {
		[14752] = {minspellid=14752,level=30, slot="SPIRIT"}, -- Divine Spirit 1
		[14818] = {minspellid=14752,level=40, slot="SPIRIT"}, -- Divine Spirit 2
		[14819] = {minspellid=14752,level=50, slot="SPIRIT"}, -- Divine Spirit 3
		[27841] = {minspellid=14752,level=60, slot="SPIRIT"}, -- Divine Spirit 4
		[27681] = {minspellid=27681,level=60, slot="SPIRIT"}, -- Prayer of Spirit 1
		[25312] = {minspellid=14752,level=70, slot="SPIRIT"}, -- Divine Spirit 5
		[32999] = {minspellid=27681,level=70, slot="SPIRIT"}, -- Prayer of Spirit 2
		[48073] = {minspellid=14752,level=80, slot="SPIRIT"}, -- Divine Spirit 6
		[48074] = {minspellid=27681,level=80, slot="SPIRIT"}, -- Prayer of Spirit 3
	},
	["WARLOCK_ARMOR"] = {
		[687] = {minspellid=687,level=1, slot="WARLOCK_ARMOR"}, -- Demon Skin 1
		[696] = {minspellid=687,level=10, slot="WARLOCK_ARMOR"}, -- Demon Skin 2
		[706] = {minspellid=706,level=20, slot="WARLOCK_ARMOR"}, -- Demon Armor 1
		[1086] = {minspellid=706,level=30, slot="WARLOCK_ARMOR"}, -- Demon Armor 2
		[11733] = {minspellid=706,level=40, slot="WARLOCK_ARMOR"}, -- Demon Armor 3
		[11734] = {minspellid=706,level=50, slot="WARLOCK_ARMOR"}, -- Demon Armor 4
		[11735] = {minspellid=706,level=60, slot="WARLOCK_ARMOR"}, -- Demon Armor 5
		[27260] = {minspellid=706,level=70, slot="WARLOCK_ARMOR"}, -- Demon Armor 6
		[47793] = {minspellid=706,level=76, slot="WARLOCK_ARMOR"}, -- Demon Armor 7
		[47889] = {minspellid=706,level=80, slot="WARLOCK_ARMOR"}, -- Demon Armor 8
		[28176] = {minspellid=28176,level=62, slot="WARLOCK_ARMOR"}, -- Fel Armor 1
		[28189] = {minspellid=28176,level=69, slot="WARLOCK_ARMOR"}, -- Fel Armor 2
		[47892] = {minspellid=28176,level=74, slot="WARLOCK_ARMOR"}, -- Fel Armor 3
		[47893] = {minspellid=28176,level=79, slot="WARLOCK_ARMOR"}, -- Fel Armor 4
	},
	["MAGE_ARMOR"] = {
		[6117] = {minspellid=6117,level=34, slot="MAGE_ARMOR"}, -- Mage Armor 1
		[22782] = {minspellid=6117,level=46, slot="MAGE_ARMOR"}, -- Mage Armor 2
		[22783] = {minspellid=6117,level=58, slot="MAGE_ARMOR"}, -- Mage Armor 3
		[27125] = {minspellid=6117,level=69, slot="MAGE_ARMOR"}, -- Mage Armor 4
		[43023] = {minspellid=6117,level=71, slot="MAGE_ARMOR"}, -- Mage Armor 5
		[43024] = {minspellid=6117,level=79, slot="MAGE_ARMOR"}, -- Mage Armor 6
		[30482] = {minspellid=30482,level=62, slot="MAGE_ARMOR"}, -- Molten Armor 1
		[43045] = {minspellid=30482,level=71, slot="MAGE_ARMOR"}, -- Molten Armor 2
		[43046] = {minspellid=30482,level=79, slot="MAGE_ARMOR"}, -- Molten Armor 3
		[168] = {minspellid=168,level=1, slot="MAGE_ARMOR"}, -- Frost Armor 1
		[7300] = {minspellid=168,level=10, slot="MAGE_ARMOR"}, -- Frost Armor 2
		[7301] = {minspellid=168,level=20, slot="MAGE_ARMOR"}, -- Frost Armor 3
		[7302] = {minspellid=7302,level=30, slot="MAGE_ARMOR"}, -- Ice Armor 1
		[7320] = {minspellid=7302,level=40, slot="MAGE_ARMOR"}, -- Ice Armor 2
		[10219] = {minspellid=7302,level=50, slot="MAGE_ARMOR"}, -- Ice Armor 3
		[10220] = {minspellid=7302,level=60, slot="MAGE_ARMOR"}, -- Ice Armor 4
		[27124] = {minspellid=7302,level=69, slot="MAGE_ARMOR"}, -- Ice Armor 5
		[43008] = {minspellid=7302,level=79, slot="MAGE_ARMOR"}, -- Ice Armor 6
	},
	["PALADIN_SEAL"] = {
		[53736] = {minspellid=53736,level=66, slot="SEALS"}, -- seal of corruption
		[20164] = {minspellid=20164,level=22, slot="SEALS"}, -- seal of justice
		[20165] = {minspellid=20165,level=30, slot="SEALS"}, -- seal of light
		[20154] = {minspellid=20154,level=1, slot="SEALS"}, -- seal of righteousness
		[31801] = {minspellid=31801,level=64, slot="SEALS"}, -- seal of vengeance
		[20166] = {minspellid=20166,level=38, slot="SEALS"}, -- seal of wisdom
		[20375] = {minspellid=20375,level=20, slot="SEALS"}, -- seal of command
	},
	["RIGHTEOUS_FURY"] = {
		[25780] = {minspellid=25780,level=20, slot="RIGHTEOUS_FURY"}, -- Righteous Fury
	},
	["PRIEST_FORM"] = {
		[15473] = {minspellid=15473,level=40, slot="PRIEST_FORM"}, -- Shadow Form
	},

	["OWL_FORM"] = {
		[24858] = {minspellid=24858,level=40, slot="SHAPESHIFT"}, -- Owl Form
	},

	["TREE_FORM"] = {
		[33891] = {minspellid=33891,level=50, slot="SHAPESHIFT"}, -- Tree Form
	},
	["PRIEST_INNER"] = {
		[588] = {minspellid=588,level=12, slot="PRIEST_INNER"}, -- Inner Fire 1
		[7128] = {minspellid=588,level=20, slot="PRIEST_INNER"}, -- Inner Fire 2
		[602] = {minspellid=588,level=30, slot="PRIEST_INNER"}, -- Inner Fire 3
		[1006] = {minspellid=588,level=40, slot="PRIEST_INNER"}, -- Inner Fire 4
		[10951] = {minspellid=588,level=50, slot="PRIEST_INNER"}, -- Inner Fire 5
		[10952] = {minspellid=588,level=60, slot="PRIEST_INNER"}, -- Inner Fire 6
		[25431] = {minspellid=588,level=69, slot="PRIEST_INNER"}, -- Inner Fire 7
		[48040] = {minspellid=588,level=71, slot="PRIEST_INNER"}, -- Inner Fire 8
		[48168] = {minspellid=588,level=77, slot="PRIEST_INNER"}, -- Inner Fire 9
	},
	["EMBRACE"] = {
		[15286] = {minspellid=15286,level=30, slot="EMBRACE"}, -- Vampiric Embrace
	},
}


BBConstants.wotlk.BuffPriorities = {
	[BBConstants.wotlk.Roles.None] = {},
	[BBConstants.wotlk.Roles.Unknown] = {},
	[BBConstants.wotlk.Roles.ManaUser] = {
		["STAMINA"] = 4,
		["WILD"] = 3,
		["SPIRIT"] = 2,
		["INTELLECT"] = 1,
	},
	[BBConstants.wotlk.Roles.NoMana] = {
		["STAMINA"] = 2,
		["WILD"] = 1,
	},
	[BBConstants.wotlk.Roles.Tank] = {
		["PALADIN_SEAL"] = 11,
		["TANK"] = 10,
		["HP"] = 9,
		["STATS"] = 8,
		["STAMINA"] = 7,
		["WILD"] = 6,
		["AP"] = 5,
		["THORNS"] = 4,
		["STRENGTH_AGILITY"] = 3,
		["MELEE_HASTE"] = 2,
		["ARMOR"] = 1,
		["SHADOW_RESIST"] = 0
	},
	[BBConstants.wotlk.Roles.PureMelee] = {
		["AP"] = 7,
		["STATS"] = 6,
		["HP"] = 5,
		["WILD"] = 4,
		["STAMINA"] = 3,
		["STRENGTH_AGILITY"] = 2,
		["MELEE_HASTE"] = 1,
		["SHADOW_RESIST"] = 0
	},
	[BBConstants.wotlk.Roles.MeleeMana] = {
		["PALADIN_SEAL"] = 13,
		["AP"] = 12,
		["STATS"] = 11,
		["MP5"] = 10,
		["HP"] = 9,
		["WILD"] = 8,
		["INTELLECT"] = 7,
		["STAMINA"] = 6,
		["STRENGTH_AGILITY"] = 5,
		["MELEE_HASTE"] = 4,
		["FLAT_SPELL_CRIT"] = 3,
		["FLAT_SPELL"] = 2,
		["SPELL_HASTE"] = 1,
		["SHADOW_RESIST"] = 0},
	[BBConstants.wotlk.Roles.Caster] = {
		["EMBRACE"] = 13,
		["PALADIN_SEAL"] = 12,
		["STATS"] = 11,
		["MP5"] = 10,
		["HP"] = 9,
		["WILD"] = 8,
		["INTELLECT"] = 7,
		["SPIRIT"] = 6,
		["STAMINA"] = 5,
		["FLAT_SPELL_CRIT"] = 4,
		["FLAT_SPELL"] = 3,
		["SPELL_HASTE"] = 2,
		["SHADOW_RESIST"] = 1,
	},
	[BBConstants.wotlk.Roles.MeleeMana_Low] = {
		["PALADIN_SEAL"] = 13,
		["MP5"] = 12,
		["AP"] = 11,
		["STATS"] = 10,
		["HP"] = 9,
		["WILD"] = 8,
		["INTELLECT"] = 7,
		["STAMINA"] = 6,
		["STRENGTH_AGILITY"] = 5,
		["MELEE_HASTE"] = 4,
		["FLAT_SPELL_CRIT"] = 3,
		["FLAT_SPELL"] = 2,
		["SPELL_HASTE"] = 1,
		["SHADOW_RESIST"] = 0
	},
	[BBConstants.wotlk.Roles.Healer] = {
		["EMBRACE"] = 12,
		["PALADIN_SEAL"] = 11,
		["MP5"] = 10,
		["STATS"] = 9,
		["HP"] = 8,
		["WILD"] = 7,
		["INTELLECT"] = 6,
		["SPIRIT"] = 5,
		["STAMINA"] = 4,
		["FLAT_SPELL_CRIT"] = 3,
		["FLAT_SPELL"] = 2,
		["SPELL_HASTE"] = 1,
		["SHADOW_RESIST"] = 0
	},
}

BBConstants.wotlk.classDefaults = {
	[BBConstants.wotlk.Classes.Warrior] = {
		profile = {
			slots = {
				[1] = {
					["SHOUTS"] = true,
				},
				[2] = {
					["SHOUTS"] = true,
				},
			},
		},
	},
	[BBConstants.wotlk.Classes.Deathknight] = {
		profile = {
			slots = {
				[1] = {
					["HORNS"] = true,
				},
				[2] = {
					["HORNS"] = true,
				},
			},
		},
	},
	[BBConstants.wotlk.Classes.Paladin] = {
		profile = {
			refreshAtDuration = 300,
			raidBuffAttendThreshold = 90,
			friendly = true,
			slots = {
				[1] = {
					["BLESSINGS"] = true,
					["SEALS"] = true,
					["RIGHTEOUS_FURY"] = false, -- not quite working yet
					--["AURAS"] = true,
				},
				[2] = {
					["BLESSINGS"] = true,
					["SEALS"] = true,
					["RIGHTEOUS_FURY"] = false, -- not quite working yet
					--["AURAS"] = true,
				},
			},
		},
	},
	[BBConstants.wotlk.Classes.Shaman] = {
		profile = {
			refreshAtDuration = 300,
			friendly = true,
			slots = {
				[1] = {
					["WATER_TOTEMS"] = true,
					["EARTH_TOTEMS"] = true,
					["FIRE_TOTEMS"] = true,
					["AIR_TOTEMS"] = true,
					-- ["SHAMAN_WEAPONS"] = true,
				},
				[2] = {
					["WATER_TOTEMS"] = true,
					["EARTH_TOTEMS"] = true,
					["FIRE_TOTEMS"] = true,
					["AIR_TOTEMS"] = true,
					-- ["SHAMAN_WEAPONS"] = true,
				},
			},
		},
	},
	[BBConstants.wotlk.Classes.Hunter] = {
		profile = {
			slots = {
				[1] = {
				},
				[2] = {
				},
			},
		},
	},
	[BBConstants.wotlk.Classes.Rogue] = {
		profile = {
			slots = {
				[1] = {
					-- ["ROGUE_POISONS"] = true,
				},
				[2] = {
					-- ["ROGUE_POISONS"] = true,
				},
			},
		},
	},
	[BBConstants.wotlk.Classes.Druid] = {
		profile = {
			refreshAtDuration = 300,
			raidBuffAttendThreshold = 90,
			raidBuffHeadCountThreshold = 2,
			slots = {
				[1] = {
					["THORNS"] = true,
					["WILD"] = true,
					["SHAPESHIFT"] = false,
				},
				[2] = {
					["THORNS"] = true,
					["WILD"] = true,
					["SHAPESHIFT"] = false,
				},
			},
		},
	},
	[BBConstants.wotlk.Classes.Warlock] = {
		profile = {
			slots = {
				[1] = {
					["WARLOCK_ARMOR"] = true,
					--["WARLOCK_WEAPON"] = true,
				},
				[2] = {
					["WARLOCK_ARMOR"] = true,
					--["WARLOCK_WEAPON"] = true,
				},
			},
		},
	},
	[BBConstants.wotlk.Classes.Priest] = {
		profile = {
			refreshAtDuration = 300,
			raidBuffAttendThreshold = 90,
			raidBuffHeadCountThreshold = 2,
			slots = {
				[1] = {
					["FORTITUDE"] = true,
					["SPIRIT"] = true,
					["PRIEST_INNER"] = true,
					["PRIEST_PROTECTION"] = true,
					["EMBRACE"] = true,
					["PRIEST_FORM"] = true,
				},
				[2] = {
					["FORTITUDE"] = true,
					["SPIRIT"] = true,
					["PRIEST_INNER"] = true,
					["PRIEST_PROTECTION"] = true,
					["EMBRACE"] = true,
					["PRIEST_FORM"] = true,
				},
			},
		},
	},
	[BBConstants.wotlk.Classes.Mage] = {
		profile = {
			refreshAtDuration = 300,
			raidBuffAttendThreshold = 90,
			raidBuffHeadCountThreshold = 2,
			slots = {
				[1] = {
					["AI"] = true,
					["MAGE_ARMOR"] = true,
					--["MAGE_STONE"] = true,
				},
				[2] = {
					["AI"] = true,
					["MAGE_ARMOR"] = true,
					--["MAGE_STONE"] = true,
				},
			},
		},
	},
}
